/* clwp 1.3: clwp.c
   C code for the co-operative lwp package,
   C.M. (Kees) Moerman, June 2002 - Januari 2003 (c)
   For documentation, see file clwp.txt
   Based on: lwp 2.0 of Josh Turpen

   Compile with -DDIAG to include error messages to standard output

   Version 1.0: changed to co-operative, added clwpDelay,
                free list, diagnostics routines
   Version 1.1: error messages enhanced, clwpLastError, clwpExit
                bug in stack handling in lwpKill (free stack???),
   Version 1.2: example4 improved
   Version 1.3 Januari 2003, C.M. Moerman (c)
                Added clwpThrow and clwpCatch
                clwpLastError is now depriciated.
                More security checks added
                All references to assert() removed
                Bug of original LWP in clwpDeleteSemaphore solved
                clwpGetThreadPriority, clwpAdjustGetThreadPriority ifx changed
*/

/* #define DIAG 1 */    /* enable for diagnostic messages etc */

#include "clwp.h"
#include <string.h>
#include <stdlib.h>     
#include <time.h>       /* used in clwpDelay() */

#ifdef DIAG
#include <stdio.h>
#endif /* DIAG */

#define CLWP_MAIN -1                    /* Main thread's PID */

/* As bugs in multi-threaded app' might be difficult to find, I
   have special markers in the data structures to check for damage.
   For example, on the maximum top of stack I place a CLWP_STACK
   marker, if this is gone the stack has overflowed (and thus run
   out of its allocated memory, possibly damaging other data). When
   switching away from a thread, this is checked by clwpYield,
   and a possible error is thrown bt clwpThrow. */

#define CLWP_CHECK 0x12345678           /* data structure marker */
#define CLWP_STACK 0x5a5a5a5a           /* stack overflow marker */
#define CLWP_START 0xdeadabba           /* stack underflow marker */

/* don't change this structure without adapting the assembly code!!!! */
typedef struct clwpStruct
{
            int check;                  /* +0:  CLWP_CHECK value for validity */
         struct clwpStruct *next;       /* +4:  Pointer to next struct */
            int stklen;                 /* +8:  Size of stack for this clwp */
       unsigned *stackptr;              /* +12: Pointer to the stack */
            int priority;               /* +16: Priority of clwp */
            int pcount;                 /* +20: Used by scheduler */
            int active;                 /* +0x18: Thread suspend */
       unsigned *stack;                 /* +28: Start of stack (low mem) */
            int clwpid;                 /* +32: Process ID */        
} clwp;

typedef struct semaphoreStruct
{           int owner;                  /* Process ID that created it */
            int clwpidLock;             /* Pid that did the lock */
            int lock;                   /* Lock value */
            int countMax;               /* Number of threads allowed */
            int count;                  /* Number of threads using */
           void *lockaddr;              /* Address of locked func/var */
         struct semaphoreStruct *next;  /* Pointer to next struct */
} clwpSemaphore;

/* ---------- below variables also used from assembly code !! ---------- */

char _clwpFpuState[108];                /* Default FPU state (27 words) for thread */
int  _clwpFlags;                        /* Default eflags for clwp */
int  _clwpDs;                           /* default ds for clwp (not used anymore?) */

clwp *_clwpCur = NULL;                  /* The running thread's struct */
clwp *_clwpMain = NULL;                 /* The main thread (used only in clwpThrow) */

int  clwpLastError = 0;                 /* global variable containing last error number */
int _clwpLastThread = -1;               /* global: thread generating error */

volatile int _clwpCount;                /* Number of running threads */
static int   _clwpPidCount;             /* Serial number for PID's */

clwp *_clwpFreeList;                    /* released thread space list */

/* ---------- above  variables also used from assembly code !! ----------- */

/* ---------- Some function prototypes used in the assembly code ---------- */

extern void _clwpInitFpu(void);         /* Init's _fpu_state */
extern int  _clwpInitFlags();           /* Gets default flags */
extern int  _clwpInitDs();              /* Gets default ds */
extern void _clwpDeadYield();           /* jik function returns */
extern void _clwpThrowAss();            /* Yield to main */


/* Implementation */

int clwpOperatingMode = CLWPMODE_STACKSIZE; /* debug/diagnostic mode */

clwpSemaphore *_clwpSemaphoreList;      /* Head of semaphore List */
/* clwpSemaphore *_clwpFreeSemaList;    (* released semaphore list */

static int _stackcheck = CLWP_STACK;    /* stack marker for main thread check */



static void clwpSetError(int errorcode) /* set error and thread id */
{
        clwpLastError = errorcode;
        _clwpLastThread = (_clwpCur != NULL)?(_clwpCur->clwpid):CLWP_MAIN;  /* Getpid() */
}

int clwpInit(int priority)
{ 
        if(priority == 0)
        {
            #ifdef DIAG
                printf("(clwpInit) Priority may not be zero.\n");
            #endif
            clwpSetError(CLWPerr_priozero);
            return(FALSE);
        }
        
        if(atexit(clwpExit))
        {
            #ifdef DIAG
                printf("(clwpInit) Couldn't install atexit routine.\n");
            #endif
            clwpSetError(CLWPerr_cantinstallatexit);
            return(FALSE);
        }

        if((_clwpCur = (clwp *) malloc(sizeof(clwp))) == NULL)
        {
            #ifdef DIAG                 /* struct for main thread */
                printf("(clwpInit) Couldn't malloc clwp structure\n");
            #endif
            clwpSetError(CLWPerr_initcantmalloc);
            return(FALSE);
        }
        _clwpMain = _clwpCur;           /* used in Throw to get back to main */
        
        _clwpCur->clwpid = CLWP_MAIN;   /* initialise main program thread struct */
        _clwpCur->next   = _clwpCur;    /* circular list */
        _clwpCur->stack  = &_stackcheck;  /* tricked only for main thread */      
        _clwpCur->priority = priority;
        _clwpCur->pcount = 1;
        _clwpCur->active = TRUE;
        _clwpCur->check  = CLWP_CHECK;  /* check marker: structure still valid? */
        
        _clwpInitFpu();                 /* get image of initial FPU state */
        _clwpFlags       = _clwpInitFlags();
        _clwpDs          = _clwpInitDs();
        _clwpCount       = 0;
        _clwpPidCount    = 0;

        _clwpSemaphoreList = NULL;        
/*      _clwpFreeSemaList = NULL;       (* released semaphore list */
        _clwpFreeList = NULL;           /* released thread space list */

        return(TRUE);                   /* success at last! */
}

int clwpGetpid(void)                    /* get process id of current thread */
{  
        if(_clwpCur == NULL) { clwpSetError(CLWPerr_clwpnoinit); return 0; }
        if(_clwpCur->check != CLWP_CHECK) clwpThrow(CLWPerr_damaged);
        return(_clwpCur->clwpid);
}

int clwpThreadCount(void)               /* number of running threads */
{
        return(_clwpCount);
}

static clwp * _clwpGetThreadPtr(int clwpid, /*@unused@*/ char *diag)
{                                       /* given pid, find pointer to struct */
        clwp *tmpclwpPtr;
        int i = 0;

        tmpclwpPtr = _clwpCur;          /* go follow list searching for pid */
        while((i<_clwpCount) && (tmpclwpPtr->clwpid != clwpid))
        {
            i++;
/* code for stack overflow analysis: overflow is serious, can't continue */
            if (tmpclwpPtr->stack[0] != CLWP_STACK) /* marker still present? */
                clwpThrow(CLWPerr_stackoverflow);
            tmpclwpPtr = tmpclwpPtr->next;
        }
        if(tmpclwpPtr->clwpid != clwpid) /* done complete circle, still not found? */
        {
            #ifdef DIAG
                printf("(clwp%s) PID not found\n", diag);
            #endif
            return(NULL);
        }
        return(tmpclwpPtr);
}

int clwpGetThreadPriority(int clwpid)
{
        clwp *tmpclwpPtr;
        tmpclwpPtr = _clwpGetThreadPtr(clwpid, "GetThreadPriority");
        if( tmpclwpPtr == NULL) return 0;
        return(tmpclwpPtr->priority);
}

int clwpAdjustThreadPriority(int clwpid, int priority)
{       int old_prio;
        clwp *tmpclwpPtr;
        tmpclwpPtr = _clwpGetThreadPtr(clwpid, "AdjustThreadPriority");
        if( tmpclwpPtr == NULL) return 0;

        old_prio = tmpclwpPtr->priority;
        tmpclwpPtr->priority = priority;
        return(old_prio);
}

/* spawns a light weight process.  Returns the clwpid of the proc or 0 on fail*/

int clwpSpawn(void (*proc)(void *), void *param, int stackLength, int priority, int active)
{
        clwp *tmpclwpPtr;               /* spawn a new thread */
        clwp *tmpclwpPtr2 = NULL;

        if(_clwpCur == NULL) { clwpSetError(CLWPerr_clwpnoinit); return 0; }
        tmpclwpPtr = _clwpFreeList;
        if((proc == NULL) || (stackLength < 1024))
        {
            #ifdef DIAG
                printf("(clwpSpawn) Attempt to spawn with a stack < 1024 bytes\n");
            #endif
            clwpSetError(CLWPerr_stacktoosmall);
            return(0);
        }

        stackLength = (stackLength + 0xF) & 0xFFFFF0; /* stack divisible by 16 */
        
        while((tmpclwpPtr != NULL) && (stackLength != tmpclwpPtr->stklen))
        {
            tmpclwpPtr2 = tmpclwpPtr;   /* check free list for matching element */
            tmpclwpPtr = tmpclwpPtr->next;
        }
        if(tmpclwpPtr != NULL)          /* found a match */
        {                               /* unlink node from list */
            if(tmpclwpPtr2 == NULL)
                _clwpFreeList = tmpclwpPtr->next;
            else
                tmpclwpPtr2->next = tmpclwpPtr->next;
        }
        else                            /* not found: malloc new memory block */
        {   tmpclwpPtr = (clwp *) malloc(sizeof(clwp));
            if(tmpclwpPtr == NULL)      /* one block for clwp struct */
            {
                #ifdef DIAG
                    printf("(clwpSpawn) Couldn't allocate memory\n");
                #endif
                clwpSetError(CLWPerr_spawncantmalloc);
                return(0);
            }
            tmpclwpPtr->stack = (unsigned *) malloc((size_t)stackLength);
            if(tmpclwpPtr->stack == NULL)
            {                           /* and one block for stack */
                #ifdef DIAG
                    printf("(clwpSpawn) Couldn't allocate stack\n");
                #endif
                clwpSetError(CLWPerr_spawncantstack);
                return(0);
            }
        }
        
/* code for stack fill with marker code to later check space used in clwpStackUsed */
        if(clwpOperatingMode & CLWPMODE_STACKSIZE)
            memset((char *)tmpclwpPtr->stack, (CLWP_STACK & 0xFF), (size_t)stackLength);
        else
            tmpclwpPtr->stack[0] = CLWP_STACK;  /* stack end marker */

        if(priority == 0) /* infinite priority! */
        {
            #ifdef DIAG
                printf("(clwpSpawn) Attempted to spawn with a priority of zero.\n"); 
            #endif
            clwpSetError(CLWPerr_priozero);
            return(0);
        }

         /* Set up stack frame to be able to do a 'pop all; ret;' in clwpYield */
        tmpclwpPtr->stack[(stackLength/4) - 1] = CLWP_START;   /* security marker */
        tmpclwpPtr->stack[(stackLength/4) - 2] = (int) param;
        tmpclwpPtr->stack[(stackLength/4) - 3] = (int) _clwpDeadYield;
        tmpclwpPtr->stack[(stackLength/4) - 4] = (int) proc;
        tmpclwpPtr->stack[(stackLength/4) - 5] = 0; /* eax */
        tmpclwpPtr->stack[(stackLength/4) - 6] = 0; /* ebx */
        tmpclwpPtr->stack[(stackLength/4) - 7] = 0; /* ecx */ 
        tmpclwpPtr->stack[(stackLength/4) - 8] = 0; /* edx */
        tmpclwpPtr->stack[(stackLength/4) - 9] = 0; /* ebp */
        tmpclwpPtr->stack[(stackLength/4) -10] = 0; /* esp, discarded by popal */
        tmpclwpPtr->stack[(stackLength/4) -11] = 0; /* esi */
        tmpclwpPtr->stack[(stackLength/4) -12] = 0; /* edi */
        tmpclwpPtr->stack[(stackLength/4) -13] = _clwpFlags; /* flags */

        /* copy initial FPU state to stack */
        memcpy(tmpclwpPtr->stack + (stackLength/4) - 13 - 108/4, _clwpFpuState, 108);
        
        tmpclwpPtr->stackptr = tmpclwpPtr->stack + (stackLength/4) - 13 - 108/4;
        tmpclwpPtr->stklen   = stackLength;
        tmpclwpPtr->priority = priority;
        tmpclwpPtr->pcount   = priority;
        tmpclwpPtr->active   = active;
        if(_clwpPidCount == -2) _clwpPidCount = 0; /* can wrap..... although not likely */
        tmpclwpPtr->clwpid   = ++_clwpPidCount;
        tmpclwpPtr->next     = _clwpCur->next;
        tmpclwpPtr->check    = CLWP_CHECK;   /* to check overwrites */
        _clwpCur->next = tmpclwpPtr;         /* insert after current */
        _clwpCount++;

/* #ifdef DIAG
        printf("clwpSpawn  Id=%04x Ds=%04x, prio=%d\n",
            tmpclwpPtr->clwpid, _clwpDs, priority);
#endif
*/
        
        return(tmpclwpPtr->clwpid);
}


int clwpCatch(void)                     /* returns error code and error pid */
{
        if(clwpLastError ==0) return 0;
        return (clwpLastError & 0xFFFF) | (_clwpLastThread << 16);
}

void clwpThrow(int errcode)             /* Go throw error code */
{
        clwpSetError(errcode);          /* set err code and thread */
        if((errcode == 0) || (_clwpLastThread == CLWP_MAIN)) return;
        _clwpThrowAss();                /* and back to main program */
}                                       /* note: this function never returns normally */

int clwpCheck(void)                     /* security check */
{                                       /* checks all threads for: */
        clwp *tmpclwpPtr;               /* - stack overflow (CLWP_STACK gone) */
        int i;                          /* - damaged structs (CLWP_CHECK gone) */
        tmpclwpPtr = _clwpMain;         /* - loose pointers */
        if (tmpclwpPtr == NULL)         /* - broken circular link */
        {   clwpThrow(CLWPerr_null_ptr); return clwpCatch(); }
        for(i=0; i <= _clwpCount; i++)
        {
            if (tmpclwpPtr == NULL)
            {   clwpThrow(CLWPerr_null_ptr); return clwpCatch(); }
            if (tmpclwpPtr->check != CLWP_CHECK)
            {   clwpThrow(CLWPerr_damaged);  return clwpCatch(); }
            if (tmpclwpPtr->stack[0] != CLWP_STACK)
            {   clwpThrow(CLWPerr_null_ptr); return clwpCatch(); }
            if((tmpclwpPtr->clwpid != CLWP_MAIN)
            && (tmpclwpPtr->stack[(tmpclwpPtr->stklen/4) - 1] != CLWP_START))
            {   clwpThrow(CLWPerr_null_ptr); return clwpCatch(); }
            tmpclwpPtr = tmpclwpPtr->next;
        }
        if(tmpclwpPtr->clwpid != CLWP_MAIN) /* full circle? */
        {   clwpThrow(CLWPerr_circular);  return clwpCatch(); }
        return 0;
}

void clwpDump()
{
#ifdef DIAG     /* only when diagnostics asked for */
        clwp *tmpclwpPtr;
        int i = 0, j = 0;
        
        tmpclwpPtr = _clwpFreeList;     /* show freed space */
        printf("Free thread list: (ds=%08x)\n", _clwpDs);
        while(tmpclwpPtr != NULL)
        {
            printf("old id=%d(%08x, stack %08x[%d])    ",
                tmpclwpPtr->clwpid, (int)tmpclwpPtr, (int)tmpclwpPtr->stack, tmpclwpPtr->stklen);
            tmpclwpPtr = tmpclwpPtr->next;
        }
        printf("\n\n");
        
        tmpclwpPtr = _clwpCur;
        while((i<=_clwpCount))          /* shows actual threads */
        {
            printf("cliwDump id=%d, active=%d, prio=%d(%d), stack=%08x[%d]\n",
                tmpclwpPtr->clwpid, tmpclwpPtr->active,
                tmpclwpPtr->priority, tmpclwpPtr->pcount,
                (int)tmpclwpPtr->stack, tmpclwpPtr->stklen 
                );
            if(tmpclwpPtr->stack)
            {   for(j=0;j<10;j++)
                {
                    printf("%03d:%08x %03d:%08x %03d:%08x %03d:%08x\n",
                    4*(j*4), tmpclwpPtr->stack[j*4],
                    4*(j*4+1), tmpclwpPtr->stack[j*4+1],
                    4*(j*4+2), tmpclwpPtr->stack[j*4+2],
                    4*(j*4+3), tmpclwpPtr->stack[j*4+3]);
                }
            }
            i++;
            tmpclwpPtr = tmpclwpPtr->next;
        }
#endif
}


int clwpStackUsed(int clwpid)           /* calc and return max. bytes used */
{
        clwp *tmpclwpPtr;
        int i = 0;
        
        if( !(clwpOperatingMode & CLWPMODE_STACKSIZE)) return 0;
        if(clwpid == CLWP_MAIN)
        {
            clwpSetError(CLWPerr_stackmain);
            #ifdef DIAG
                printf("(clwpStackUsed) can't analyse MAIN stack\n");
            #endif
            return 0;
        }

        tmpclwpPtr = _clwpGetThreadPtr(clwpid, "StackUsed");
        if(tmpclwpPtr == NULL) return 0; /* find ptr to clwpid passed */

        if (tmpclwpPtr->stack[1] != CLWP_STACK)
        {
            clwpSetError(CLWPerr_stacksize);
            #ifdef DIAG
                printf("(clwpStackUsed) Depth analysis not enabled\n");
            #endif
            return 0;
        }
        
        for(i=0; i<tmpclwpPtr->stklen/4; i++)
        {                               /* go search stack for highest altered element */
            if(tmpclwpPtr->stack[i] != CLWP_STACK)
                 return(tmpclwpPtr->stklen - i*4);
        }
        #ifdef DIAG
            printf("(clwpStackUsed) End of stack not found\n");
        #endif
        clwpThrow(CLWPerr_stackoverflow); /* should not be possible; system error */
        return 0;
}

  
/* Kills a clwp (takes it out of the linked list of clwp's) , or 0 if it fails */

int clwpKill(int clwpid)                /* kill indicated thread */
{
        int i = 0;
        clwp *tmpclwpPtr1,*tmpclwpPtr2;

        if(_clwpCur == NULL)            /* is clwp initialised? */
        {    clwpSetError(CLWPerr_clwpnoinit);
             return 0;
        }
        tmpclwpPtr1 = _clwpCur->next;
        tmpclwpPtr2 = _clwpCur;

        if(clwpid == CLWP_MAIN)
        {
            #ifdef DIAG
                printf("(clwpKill) Attempted to kill main()!\n");
            #endif
            clwpSetError(CLWPerr_cantkillmain);
            return(FALSE);
        }
        if(clwpid == 0)
        {                               /* No such thing as pid 0.  Pid's start at 1 */
            #ifdef DIAG
                printf("(clwpKill) Attempted to kill pid 0.\n");
            #endif
            clwpSetError(CLWPerr_cantkillzero);
            return(FALSE);
        }

        /* special case for the current thread */
        if(clwpid == _clwpCur->clwpid)
        {
            _clwpCur->check = -1;       /* invalidate check value */
            _clwpDeadYield();
            clwpThrow(CLWPerr_cantkill);
            #ifdef DIAG
                printf("(clwpKill) PANIC!\n"); /* should never get here after the yield */
            #endif
            exit(-1);
        }

        if((tmpclwpPtr1 == NULL) || (tmpclwpPtr2 == NULL))
        {
            #ifdef DIAG
                printf("(clwpKill) clwp list head is NULL!\n");
            #endif
            clwpThrow(CLWPerr_listheadzero);
            return(FALSE);
        }
        
        while((i< _clwpCount+1) && (tmpclwpPtr1->clwpid != clwpid))
        {
            i++;
            tmpclwpPtr1 = tmpclwpPtr1->next;
            tmpclwpPtr2 = tmpclwpPtr2->next;
        }

        if(tmpclwpPtr1->clwpid != clwpid) /* went through all of them, not found */
        {
            #ifdef DIAG
                printf("(clwpKill) PID not found\n");
            #endif
            clwpSetError(CLWPerr_pidnotfound);
            return(FALSE);
        }

        /* now tmp IS the right one to kill */
        tmpclwpPtr2->next = tmpclwpPtr1->next;
        tmpclwpPtr1->check = -1;       /* invalidate check value */
        /* free(tmpclwpPtr1->stack); free(tmpclwpPtr1); */
        tmpclwpPtr1->next = _clwpFreeList; /* add to linked free list */
        _clwpFreeList = tmpclwpPtr1;
        _clwpCount--;
        return(TRUE);
}

void clwpDelay(int seconds)             /* delay seconds by yielding */
{
        time_t start_time;

        if(_clwpCur == NULL)
        {    clwpSetError(CLWPerr_clwpnoinit);
             return;
        }
        if(seconds <= 0) return;
        start_time = time(NULL);
        while( difftime(time(NULL), start_time) < seconds)
        {
            clwpYield();
            /* printf("Time --- %d\n", (int)difftime(time(NULL), start_time)); */
        }
}

int clwpThreadSuspend(int clwpid)
{
        clwp *tmpclwpPtr;
        if(clwpid == CLWP_MAIN)
        {
            #ifdef DIAG
                printf("(clwpThreadSuspend) Tried to suspend MAIN!\n");        
            #endif
            clwpSetError(CLWPerr_suspendmain);
            return(FALSE);
        }
        tmpclwpPtr = _clwpGetThreadPtr(clwpid, "ThreadSuspend");
        if( tmpclwpPtr == NULL) return FALSE;
        tmpclwpPtr->active = FALSE;
        return(TRUE);
}

int clwpThreadResume(int clwpid)
{
        clwp *tmpclwpPtr;
        tmpclwpPtr = _clwpGetThreadPtr(clwpid, "ThreadResume");
        if( tmpclwpPtr == NULL) return FALSE;
        tmpclwpPtr->active = TRUE;
        return(TRUE);
}

clwpSemaphore * _clwpSemaInit(void *lockaddr, int count)
{
        clwpSemaphore *newSema;         /* request and initialise new Semaphore */
        newSema = (clwpSemaphore *) malloc(sizeof(struct semaphoreStruct));
        if(newSema == NULL)             /* failed to get memory? */
        {
            #ifdef DIAG
                printf("(clwpCreateSemaphore) Couldn't allocate memory\n");
            #endif
            clwpSetError(CLWPerr_semacantmalloc);
            return(NULL);
        }                               /* fill in clwpSemaphore data struct */
        newSema->lockaddr = lockaddr;
        newSema->countMax = count;
        newSema->next = NULL;
        newSema->lock = FALSE;
        newSema->clwpidLock = 0;
        newSema->owner = clwpGetpid();
        newSema->count = 0;
        return newSema;
}

int clwpCreateSemaphore(void *lockaddr, int count)
{
        clwpSemaphore *tmpclwpSemaphore;

        if(_clwpSemaphoreList == NULL)  /* first semaphore? */
        {
            tmpclwpSemaphore = _clwpSemaInit(lockaddr, count);
            if(tmpclwpSemaphore == NULL) return FALSE;
            
            _clwpSemaphoreList = tmpclwpSemaphore;

            return(TRUE);
        }

        /* Look through the list and see if it's already been created */
        tmpclwpSemaphore = _clwpSemaphoreList;
        while(tmpclwpSemaphore->lockaddr != lockaddr)
        {
            tmpclwpSemaphore = tmpclwpSemaphore->next;
            if(tmpclwpSemaphore == NULL)
            {
                /* If we've gotten this far, it hasn't been created */
                tmpclwpSemaphore = _clwpSemaInit(lockaddr, count);
                if(tmpclwpSemaphore == NULL) return FALSE;
                
                tmpclwpSemaphore->next = _clwpSemaphoreList->next;
                _clwpSemaphoreList->next = tmpclwpSemaphore;
                return(TRUE);
            }
        }

        #ifdef DIAG
            printf("(clwpCreateSemaphore) Semaphore already created\n");
        #endif
        clwpSetError(CLWPerr_semaexists);
        return(FALSE);
}

int clwpDeleteSemaphore(void *lockaddr)
{
        clwp *tmpclwpPtr;
        clwpSemaphore *tmpclwpSemaphore, *tmpclwpSemaphorePrev;
        int i;
        
        if(_clwpSemaphoreList == NULL)
        {
            #ifdef DIAG
                printf("(clwpDeleteSemaphore) Semaphore List is NULL\n");
            #endif
            clwpSetError(CLWPerr_semalistzero);
            return(FALSE);
        }

        tmpclwpSemaphore = _clwpSemaphoreList;     /* search for previous */   
        tmpclwpSemaphorePrev = _clwpSemaphoreList; /* to update list correctly */
        while(tmpclwpSemaphore->lockaddr != lockaddr)
        {
            tmpclwpSemaphorePrev = tmpclwpSemaphore;
            tmpclwpSemaphore = tmpclwpSemaphore->next;
            if(tmpclwpSemaphore == NULL)
            {
                #ifdef DIAG
                    printf("(clwpDeleteSemaphore) Semaphore does not exist\n");        
                #endif
                clwpSetError(CLWPerr_semanotexist);
                return(FALSE);
            }
        }

        /* If the semaphore owner doesn't exist, we can delete it */
        tmpclwpPtr = _clwpCur;
        i = 0;
        while((i < _clwpCount+1) && (tmpclwpPtr->clwpid != tmpclwpSemaphore->owner))
        {
            tmpclwpPtr = tmpclwpPtr->next;
            i++;
        }

        if(i == _clwpCount+1)
        {
            /* Owner doesn't exist, so safe to delete */
            tmpclwpSemaphorePrev->next = tmpclwpSemaphore->next;
            if(tmpclwpSemaphorePrev == _clwpSemaphoreList) /* bug solve in 1.3 */
                _clwpSemaphoreList = tmpclwpSemaphore->next;
            free(tmpclwpSemaphore);
            return(TRUE);
        }
        
        if(tmpclwpSemaphore->owner != clwpGetpid())
        {
            #ifdef DIAG
                printf("(clwpDeleteSemaphore) Thread deleting a semaphore it didn't create\n");
            #endif
            clwpSetError(CLWPerr_semanotowner);
            return(FALSE);
        }

        tmpclwpSemaphorePrev->next = tmpclwpSemaphore->next;
        if(tmpclwpSemaphorePrev == _clwpSemaphoreList) /* bug solve in 1.3 */
            _clwpSemaphoreList = tmpclwpSemaphore->next;
        free(tmpclwpSemaphore);
        return(TRUE);
}

int clwpAdjustSemaphoreCount(void *lockaddr, int count)
{
        clwpSemaphore *tmpclwpSemaphore;
        tmpclwpSemaphore = _clwpSemaphoreList;

        if(tmpclwpSemaphore == NULL)
        {
            #ifdef DIAG
                printf("(clwpAdjustSemaphoreCount) Semaphore list is NULL\n");        
            #endif
            clwpSetError(CLWPerr_semalistzero);
            return(FALSE);
        }

        while(tmpclwpSemaphore->lockaddr != lockaddr)
        {
            tmpclwpSemaphore = tmpclwpSemaphore->next;
            if(tmpclwpSemaphore == NULL)
            {
                #ifdef DIAG
                    printf("(clwpAdjustSemaphoreCount) Semaphore does not exist\n");
                #endif
                clwpSetError(CLWPerr_semanotexist);
                return(FALSE);
            }
        }

        tmpclwpSemaphore->count = count;        
        return(TRUE);
}

int clwpGetSemaphoreCount(void *lockaddr)
{
        clwpSemaphore *tmpclwpSemaphore;
        tmpclwpSemaphore = _clwpSemaphoreList;

        if(tmpclwpSemaphore == NULL)
        {
            #ifdef DIAG
                printf("(clwpGetSemaphoreCount) Semaphore list is NULL\n");        
            #endif
            clwpSetError(CLWPerr_semalistzero);
            return(0);
        }

        while(tmpclwpSemaphore->lockaddr != lockaddr)
        {
            tmpclwpSemaphore = tmpclwpSemaphore->next;
            if(tmpclwpSemaphore == NULL)
            {
                #ifdef DIAG
                    printf("(clwpGetSemaphoreCount) Semaphore does not exist\n");
                #endif
                clwpSetError(CLWPerr_semanotexist);
                return(0);
            }
        }

        return(tmpclwpSemaphore->count);
}

int clwpLockSemaphore(void *lockaddr)
{
        clwpSemaphore *tmpclwpSemaphore;
        clwp *tmpclwpPtr;
        int i;
        
        tmpclwpSemaphore = _clwpSemaphoreList;
        
        if(tmpclwpSemaphore == NULL)
        {
            #ifdef DIAG
                printf("(clwpLockSemaphore) Semaphore List is NULL\n");
            #endif
            clwpSetError(CLWPerr_semalistzero);
            return(FALSE);
        }
        while(tmpclwpSemaphore->lockaddr != lockaddr)
        {
            tmpclwpSemaphore = tmpclwpSemaphore->next;
            if(tmpclwpSemaphore == NULL)
            {
                #ifdef DIAG
                    printf("clwpLockSemaphore) Semaphore does not exist\n");
                #endif
                clwpSetError(CLWPerr_semanotexist);
                return(FALSE);
            }
        }

        if(tmpclwpSemaphore->count < tmpclwpSemaphore->countMax)
        {
            tmpclwpSemaphore->lock = TRUE;
            tmpclwpSemaphore->count++;
            tmpclwpSemaphore->clwpidLock = clwpGetpid();
            return(TRUE);
        }

        /* The tricky part.  Since the semaphore is already locked, we
         * have to wait for it to unlock.  But we also have to check to
         * see if the thread that created it still exists, to avoid
         * deadlocks.  
         * 
         * Known bug:  When yielding to the other threads to allow the
         * semaphore to get unlocked, the other threads might call
         * clwpDeleteSemaphore on this semaphore!  Not good!
         */        
        while(tmpclwpSemaphore->lock == TRUE)
        {
            /* Make sure thread still exists that did the lock */
            i = 0;
            tmpclwpPtr = _clwpCur;
            while((i < _clwpCount+1) && (tmpclwpPtr->clwpid != tmpclwpSemaphore->clwpidLock))
            {
                tmpclwpPtr = tmpclwpPtr->next;
                i++;
            }

            /* Thread didn't exist, so we can lock it */
            if(i == _clwpCount+1)
            {
                tmpclwpSemaphore->count++;
                tmpclwpSemaphore->clwpidLock = clwpGetpid();
                tmpclwpSemaphore->lock = TRUE;
                return(TRUE);
            }        

            clwpYield();
        }

        if(tmpclwpSemaphore->count >= tmpclwpSemaphore->countMax)
        {
            #ifdef DIAG
                printf("(clwpLockSemaphore) Lock limit exceeded\n");
            #endif
            clwpSetError(CLWPerr_semalocklimit);
            return(FALSE);
        }

        tmpclwpSemaphore->lock = TRUE;
        tmpclwpSemaphore->count++;
        tmpclwpSemaphore->clwpidLock = clwpGetpid();

        return(TRUE);
}

int clwpReleaseSemaphore(void *lockaddr)
{
        clwpSemaphore *tmpclwpSemaphore;
        
        tmpclwpSemaphore = _clwpSemaphoreList;
        if(tmpclwpSemaphore == NULL)
        {
            #ifdef DIAG
                printf("(clwpReleaseSemaphore) Semaphore list is NULL\n");
            #endif 
            clwpSetError(CLWPerr_semalistzero);
            return(FALSE);
        }
        while(tmpclwpSemaphore->lockaddr != lockaddr)
        {
            tmpclwpSemaphore = tmpclwpSemaphore->next;
            if(tmpclwpSemaphore == NULL)
            {
                #ifdef DIAG
                    printf("(clwpReleaseSemaphore) Semaphore does not exist\n");
                #endif
                clwpSetError(CLWPerr_semanotexist);
                return(FALSE);
            }
        }
        
        tmpclwpSemaphore->lock = FALSE;
        tmpclwpSemaphore->clwpidLock = 0;
        tmpclwpSemaphore->count--;
        if(tmpclwpSemaphore->count < 0)
        {
            tmpclwpSemaphore->count = 0;
        }
        return(TRUE);
}



int clwpCreateMutex(void *lockaddr)
{
        return(clwpCreateSemaphore(lockaddr, 1));
}

int clwpDeleteMutex(void *lockaddr)
{
        return(clwpDeleteSemaphore(lockaddr));
}

int clwpLockMutex(void *lockaddr)
{
        return(clwpLockSemaphore(lockaddr));
}

int clwpReleaseMutex(void *lockaddr)
{
        return(clwpReleaseSemaphore(lockaddr));
}


void clwpExit(void)                 /* called by runtime atexit routine */
{
        clwp *tmpclwpPtr;
        int i;
        
        if(_clwpCur == NULL) return;    /* done already manually? */

        for(i=0; i<= _clwpCount; i++)   /* kill running threads */
        {
            tmpclwpPtr = _clwpCur;
            _clwpCur = _clwpCur->next;  /* except for main */
            if(tmpclwpPtr->clwpid != -1) clwpKill(tmpclwpPtr->clwpid);
        }

        while(_clwpFreeList != NULL)    /* release thread space list */
        {
            tmpclwpPtr = _clwpFreeList;
#if 0
            printf("next [%d] %08x %08x -> %08x   ",
                 tmpclwpPtr->clwpid, tmpclwpPtr->stack, tmpclwpPtr, tmpclwpPtr->next);
#endif
            _clwpFreeList = _clwpFreeList->next;

#if 0
            if(tmpclwpPtr->stack[(tmpclwpPtr->stklen/4) - 1] != CLWP_START)
            {
                
                printf(" === stack underflow..... %08x === ",
                    tmpclwpPtr->stack[(tmpclwpPtr->stklen/4) - 1]);
            }
            printf(" < free %08x ", tmpclwpPtr->stack);
            printf(" free %08x > ", tmpclwpPtr);
#endif
            if(tmpclwpPtr->stack[(tmpclwpPtr->stklen/4) - 1] != CLWP_START)
            {   clwpSetError(CLWPerr_stackoverflow);
                return;
            }

            free(tmpclwpPtr->stack);
            tmpclwpPtr->check = -1;     /* invalidate check value */
            free(tmpclwpPtr);
        }              

        _clwpCur->check = -1;           /* invalidate check value */
        free(_clwpCur);                 /* release main task clwp structure */
        _clwpCur = NULL;                /* reset initialised flag */
        return;
}

void _clwp_gen_overflow(void *param)    /* 'generate stack overflow' */
{   while(1)
    {   _clwpCur->stack[0] = -1;        /* damage stack check marker */
        clwpYield();                    /* switch to next thread */
    }
}

/* end of file */

